/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla XHTML + Voice support.
 *
 * The Initial Developer of the Original Code is IBM Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Chris Cross <xcross@us.ibm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <stdio.h>
#include "nsString.h"
#include "nsCOMPtr.h"
#include "nsComponentManagerUtils.h"
#include "nsIDocument.h"
#include "nsIDOMDocument.h"
#include "nsIDOMElement.h"
#include "nsIDOMEvent.h"
#include "nsIDOMEventTarget.h"
#include "nsIDOM3Node.h"
#include "nsIDOMNamedNodeMap.h"
#include "nsIDOMSerializer.h"
#include "nsIServiceManager.h"
#include "nsIURI.h"
#include "nsIVxmlInterpreter.h"
#include "nsIXTFElement.h"
#include "nsIXTFGenericElementWrapper.h"

#include "nsVxmlFormElement.h"

#define DEBUGTRACE
#include "debugtrace.h"

NS_IMPL_ISUPPORTS3(nsVxmlFormElement, nsIVxmlFormElement, nsIXTFGenericElement, nsIDOMEventListener)

//Used by GetScriptingInterfaces
static const nsIID sScriptingIIDs[] = {
  NS_IDOMELEMENT_IID,
  NS_IDOMEVENTTARGET_IID,
  NS_IDOM3NODE_IID
};

//*********************************************************************************************************
// nsVxmlFormElement impl

NS_HIDDEN_(nsresult)
NS_NewVxmlFormElement(nsIXTFElement **aResult)
{
  FUNCTION_CLASS_TRACE (NS_NewVxmlFormElement)

  *aResult = new nsVxmlFormElement();
  if (!*aResult)
    return NS_ERROR_OUT_OF_MEMORY;

  NS_ADDREF(*aResult);
  return NS_OK;
}

nsVxmlFormElement::nsVxmlFormElement()
  : mElement(0), mCount(0), mDoc(0), mVxmlInterpreter(0), mbDoneCreatingDOM(PR_FALSE) 
{
  FUNCTION_CLASS_TRACE (nsVxmlFormElement::nsVxmlFormElement)

	printf("\n\nWoo hoo, Hello World!!!!!!!!!!!!!!!!!!!!!!!!!!!\n");
  /* member initializers and constructor code */
}

nsVxmlFormElement::~nsVxmlFormElement()
{
  FUNCTION_CLASS_TRACE (nsVxmlFormElement::~nsVxmlFormElement)
  /* destructor code */
}

nsresult 
nsVxmlFormElement::Serialize(nsAString& aString)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::serialize )

  nsCOMPtr<nsIDOMSerializer> s = do_CreateInstance(NS_XMLSERIALIZER_CONTRACTID);
  if (!s)
  {
    return NS_ERROR_FAILURE;
  }
  s->SerializeToString(mElement, aString);
  return NS_OK;
}

void nsVxmlFormElement::init()
{
  //Get the VXMLInterpreter Service
  nsCOMPtr<nsIServiceManager> sm;
  NS_GetServiceManager(getter_AddRefs(sm));
  NS_ASSERTION(sm, "Failed to get nsIServiceManager");
 
  sm->GetServiceByContractID("@mozilla.org/vxml", NS_GET_IID(nsIVxmlInterpreter), (void **)&mVxmlInterpreter);
  NS_ASSERTION(mVxmlInterpreter, "Failed to get VxmlInterpreter");
 
  nsCOMPtr<nsIDOMNode> node = do_QueryInterface(mElement);
  NS_ASSERTION(node, "Element is not a node, we'll crash soon");

  //Cache the document holding this form element
  nsCOMPtr<nsIDOMDocument> domdoc;
  node->GetOwnerDocument(getter_AddRefs(domdoc));
  mDoc = do_QueryInterface(domdoc);

  //Get the document and base URIs
  nsCOMPtr<nsIURI> baseURI = mDoc->GetBaseURI();
  getter_AddRefs(baseURI);
  nsCOMPtr<nsIURI> docURI = mDoc->GetDocumentURI();
  getter_AddRefs(docURI);
  nsCAutoString cStrDocURI, cStrBaseURI;
  docURI->GetSpec(cStrDocURI);
  baseURI->GetSpec(cStrBaseURI);

  nsAutoString strDocURI, strBaseURI;
  CopyUTF8toUTF16(cStrDocURI, strDocURI);
  CopyUTF8toUTF16(cStrBaseURI, strBaseURI);
  mBaseURI.Assign(strBaseURI);
  mDocURI.Assign(strDocURI);

  //Get the handler id from this element's node map
  nsCOMPtr<nsIDOMNamedNodeMap> nodeMap;
  node->GetAttributes(getter_AddRefs(nodeMap));
  nsCOMPtr<nsIDOMNode> idNode;
  nodeMap->GetNamedItem(NS_LITERAL_STRING("id"), getter_AddRefs(idNode)); 
  nsAutoString handlerID;
  idNode->GetNodeValue(handlerID);
  mID.Assign(handlerID);
}

//*********************************************************************************************************
// nsIVxmlFormElement methods

/* attribute DOMString innerVXML; */
NS_IMETHODIMP 
nsVxmlFormElement::GetInnerVXML(nsAString & aInnerVXML)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::GetInnerVXML )
  return Serialize(aInnerVXML);
}

NS_IMETHODIMP 
nsVxmlFormElement::SetInnerVXML(const nsAString & aInnerVXML)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::SetInnerVXML )
  return NS_ERROR_NOT_IMPLEMENTED;
}


//*********************************************************************************************************
// nsIXTFElement methods

NS_IMETHODIMP 
nsVxmlFormElement::OnDestroyed(void) 
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::OnDestroyed )

  // release our node
  mElement = nsnull;
  return NS_OK;
}

NS_IMETHODIMP 
nsVxmlFormElement::GetElementType(PRUint32 *aElementType)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::GetElementType )

  *aElementType = nsIXTFElement::ELEMENT_TYPE_GENERIC_ELEMENT;
  return NS_OK;
}



NS_IMETHODIMP nsVxmlFormElement::GetIsAttributeHandler(PRBool *aIsAttributeHandler)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::GetIsAttributeHandler )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::GetScriptingInterfaces(PRUint32 *aCount, nsIID ***aArray)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::GetScriptingInterfaces )
  return CloneScriptingInterfaces(sScriptingIIDs, NS_ARRAY_LENGTH(sScriptingIIDs),aCount, aArray);
  //return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::WillChangeDocument(nsIDOMDocument *newDoc)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::WillChangeDocument )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::DocumentChanged(nsIDOMDocument *newDoc)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::DocumentChanged )

  if(!newDoc)
    return NS_OK;

  //Cache the document holding this form element
  mDoc = do_QueryInterface(newDoc);

  //Add listener for "DOMContentLoaded" 
  nsCOMPtr<nsIDOMEventTarget> eventTarget = do_QueryInterface(mDoc);
  NS_ASSERTION(eventTarget, "Bad document event target");
  eventTarget->AddEventListener(NS_LITERAL_STRING("DOMContentLoaded"), this, PR_TRUE);

  //  OnDOMContentLoaded();
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::WillChangeParent(nsIDOMElement *newParent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::WillChangeParent )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::ParentChanged(nsIDOMElement *newParent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::ParentChanged )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::WillInsertChild(nsIDOMNode *child, PRUint32 index)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::WillInsertChild )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::ChildInserted(nsIDOMNode *child, PRUint32 index)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::ChildInserted )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::WillAppendChild(nsIDOMNode *child)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::WillAppendChild )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::ChildAppended(nsIDOMNode *child)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::ChildAppended )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::WillRemoveChild(PRUint32 index)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::WillRemoveChild )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::ChildRemoved(PRUint32 index)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::ChildRemoved )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::WillSetAttribute(nsIAtom *name, const nsAString & newValue)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::WillSetAttribute )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::AttributeSet(nsIAtom *name, const nsAString & newValue)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::AttributeSet )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::WillRemoveAttribute(nsIAtom *name)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::WillRemoveAttribute )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::AttributeRemoved(nsIAtom *name)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::AttributeRemoved )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::BeginAddingChildren(void)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::BeginAddingChildren )
  return NS_OK;
}

NS_IMETHODIMP 
nsVxmlFormElement::DoneAddingChildren(void) 
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::DoneAddingChildren )
  return NS_OK;
}

NS_IMETHODIMP nsVxmlFormElement::CloneState(nsIDOMElement *aElement)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::CloneState )
  return NS_OK;
}


NS_IMETHODIMP 
nsVxmlFormElement::OnCreated(nsIXTFGenericElementWrapper *wrapper) 
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::OnCreated )

  wrapper->SetNotificationMask(nsIXTFElement::NOTIFY_HANDLE_DEFAULT |
                               nsIXTFElement::NOTIFY_DOCUMENT_CHANGED);
   //nsIXTFElement::NOTIFY_DONE_ADDING_CHILDREN

  nsCOMPtr<nsIDOMElement> element;
  wrapper->GetElementNode(getter_AddRefs(element));

  // It's ok to keep a weak pointer to mElement.  mElement will have an
  // owning reference to this object, so as long as we null out mElement in
  // OnDestroyed, it will always be valid.

  mElement = element;
  NS_ASSERTION(mElement, "Wrapper is not an nsIDOMElement, we'll crash soon");

  return NS_OK;
}

//*********************************************************************************************************
// nsIXTFGenericElement methods

NS_IMETHODIMP 
nsVxmlFormElement::HandleDefault(nsIDOMEvent *aEvent, PRBool *_retval) 
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::HandleDefault )

  //Don't handle any event before the DOM is ready
  if(!mbDoneCreatingDOM)
    return NS_OK;


  nsAutoString type;
  PRUint16 phase = 0;
  aEvent->GetType(type);
  aEvent->GetEventPhase(&phase);
  wprintf(L"nsVxmlFormElement::HandleDefault type: %s phase: %u\n", type.get(), phase);
  *_retval = PR_FALSE;

  //Don't handle the vxml events 
  if (type.Equals(NS_LITERAL_STRING("vxmldone")) |
      type.Equals(NS_LITERAL_STRING("nomatch")) |
      type.Equals(NS_LITERAL_STRING("noinput")))
    return NS_OK;
  else
    return Activate(aEvent);
}
 
//*********************************************************************************************************
// nsIDOMEventListener methods

NS_IMETHODIMP 
nsVxmlFormElement::HandleEvent(class nsIDOMEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::HandleEvent )

  nsAutoString s;
  aEvent->GetType(s);
  PRUint16 phase = 0;
  aEvent->GetEventPhase(&phase);
  wprintf(L"nsVxmlFormElement::HandleEvent event type: %s phase: %u\n", s.get(), phase);

  if( s == NS_LITERAL_STRING("DOMContentLoaded") )
  {
    return OnDOMContentLoaded();
  }
  else
  {
    //Don't handle any event before the DOM is ready
    if(!mbDoneCreatingDOM)
      return NS_OK;
    else
  return Activate(aEvent);
}
}

nsresult 
nsVxmlFormElement::OnDOMContentLoaded()
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::OnDOMContentLoaded )

  //The dom is created so we can initialize our member variables
  mbDoneCreatingDOM = PR_TRUE;
  init();

  //Initialize the ECMAScript scope
  mVxmlInterpreter->InitDocumentScope(mDoc);

  //Allocate the interpreter if not done already
  //todo: remove this allocate and make sure nsIVxmlInterpreter::RunHandler does it
  PRInt32 state;
  PRInt16 rv;
  mVxmlInterpreter->GetState(&state);
  if(!(state & nsIVxmlInterpreter::ALLOCATED))
     mVxmlInterpreter->Allocate(&rv);

  //Next get the text for this form element and send to the interperter
  nsAutoString text;
  nsresult rc = Serialize(text);

  mVxmlInterpreter->AddHandler(text.get(),mDocURI.get(),getTimeStamp(),mBaseURI.get(),mID.get(),&rv);

  return rc;
}

nsresult
nsVxmlFormElement::Activate(class nsIDOMEvent *aEvent)
{
   FUNCTION_CLASS_TRACE ( nsVxmlFormElement::HandleEvent )

   PRInt16 rc;
   mVxmlInterpreter->RunHandler(this, mID.get(), mDocURI.get(), false, &rc);
   return NS_OK;
}

/* static */ nsresult
nsVxmlFormElement::CloneScriptingInterfaces(const nsIID *aIIDList,
                                        unsigned int aIIDCount,
                                        PRUint32 *aOutCount,
                                        nsIID ***aOutArray)
{
  nsIID **iids = NS_STATIC_CAST(nsIID**,
                                nsMemory::Alloc(aIIDCount * sizeof(nsIID*)));
  if (!iids) {
    return NS_ERROR_OUT_OF_MEMORY;
  }
 
  for (PRUint32 i = 0; i < aIIDCount; ++i) {
    iids[i] = NS_STATIC_CAST(nsIID*,
                             nsMemory::Clone(&aIIDList[i], sizeof(nsIID)));
 
    if (!iids[i]) {
      for (PRUint32 j = 0; j < i; ++j)
        nsMemory::Free(iids[j]);
      nsMemory::Free(iids);
      return NS_ERROR_OUT_OF_MEMORY;
    }
  }
 
  *aOutArray = iids;
  *aOutCount = aIIDCount;
  return NS_OK;
}

