/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mozilla XHTML + Voice support.
 *
 * The Initial Developer of the Original Code is IBM Corporation.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *  Chris Cross <xcross@us.ibm.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include <direct.h>

#include "nsVxmlInterpreter.h"
#include "jsapi.h"
#include "nsIDocument.h"
#include "nsIDOMDocument.h"
#include "nsIDOMDocumentEvent.h"
#include "nsIDOMDocumentView.h"
#include "nsIDOMElement.h"
#include "nsIDOMEvent.h"
#include "nsIDOMEventTarget.h"
#include "nsIDOMAbstractView.h"
#include "nsIDOMNode.h"
#include "nsIJSContextStack.h"
#include "nsIObserverService.h"
#include "nsIScriptGlobalObject.h"
#include "nsIScriptContext.h"
#include "nsIServiceManager.h"
#include "nsStringAPI.h"
#include "nsISupportsPrimitives.h"
#include "nsIVxmlFormElement.h"


#include "vxmscope.h"
#include "nsMozillaScope.h"

NS_IMPL_THREADSAFE_ISUPPORTS4(nsVxmlInterpreter, nsIVxmlInterpreter, nsIDOMEventListener, nsIObserver, nsIRunnable)
NS_IMPL_ISUPPORTS1(nsVxmlFeatureDescriptor, nsIVxmlFeatureDescriptor)
NS_IMPL_ISUPPORTS1(nsVxmlPropertyDescriptor, nsIVxmlPropertyDescriptor)
NS_IMPL_ISUPPORTS1(nsVxmlLogListener, nsIVxmlLogListener)
NS_IMPL_THREADSAFE_ISUPPORTS1(RunHandlerRunnable, nsISupports)

#define DEBUGTRACE
#include "debugtrace.h"

nsVxmlInterpreter::nsVxmlInterpreter() 
  : mbStartListening(PR_FALSE), 
    mCurrentDoc(nsnull),
    mState(0),
    mStateChangeMonitor(nsnull),
    mbShuttingDown(PR_FALSE),
    mbDocUnloading(PR_FALSE),
    mbDocChanging(PR_FALSE),
    mDocChangingMonitor(nsnull),
    mEventQ(nsnull),
    mEventQThread(nsnull),
    mEventQMonitor(nsnull),
    mRunHandlerThread(nsnull)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::nsVxmlInterpreter)

  /* member initializers and constructor code */
  mLoadedDocuments.Init();
  m_pVxm = VxmInterpreter::createInstance();
  NS_ASSERTION(m_pVxm, "Memory allocation error creating Vxm.");

  //set the listeners
  VxmRc rc;
  rc = m_pVxm->addXMLEventListener(this);
  NS_ASSERTION(rc == VXM_RC_OK, "Error adding VxmXMLEventListener.");

  rc = m_pVxm->addStateListener(this);
  NS_ASSERTION(rc == VXM_RC_OK, "Error adding VxmStateListener.");

  //Initialize the State variables
  mStateChangeMonitor = PR_NewMonitor();
  GetState(&mState);
  mDocChangingMonitor = PR_NewMonitor();

  //Add Observers
  nsCOMPtr<nsIObserverService> obs(do_GetService("@mozilla.org/observer-service;1"));
  NS_ASSERTION(obs, "Unable to obtain Observer service.");
  obs->AddObserver(this, "xpcom-shutdown", PR_FALSE);

  //Start the event queue
  mEventQMonitor = PR_NewMonitor();
  nsresult rv = NS_NewThread(getter_AddRefs(mEventQThread),
                             this,
                             0,
                             PR_JOINABLE_THREAD);
  PR_Wait(mEventQMonitor, PR_INTERVAL_NO_TIMEOUT);




#if DEBUG
  //Start a thread for keyboard push-to-talk
  PRThread * pttThrd = 
    PR_CreateThread(PR_USER_THREAD, ptt, (void *)this, PR_PRIORITY_HIGH, PR_GLOBAL_THREAD, PR_JOINABLE_THREAD, 0);
#endif
}


nsVxmlInterpreter::~nsVxmlInterpreter()
{
  delete m_pVxm;
  mCurrentDoc = nsnull;
  PR_DestroyMonitor(mStateChangeMonitor);
  mStateChangeMonitor = nsnull;
  PR_DestroyMonitor(mDocChangingMonitor);
  mDocChangingMonitor = nsnull;
  /* destructor code */
}

//*********************************************************************************************************
// nsIVxmlInterpreter methods

NS_IMETHODIMP nsVxmlInterpreter::CreateInstance(nsIVxmlInterpreter **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::SetFeatureDescriptor(nsIVxmlFeatureDescriptor *pFd, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::GetFeatureDescriptor(nsIVxmlFeatureDescriptor **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::SetPropertyDescriptor(nsIVxmlPropertyDescriptor *aPD, PRInt16 *_retval)
{
  nsVxmlPropertyDescriptor * pd = NS_STATIC_CAST(nsVxmlPropertyDescriptor *, aPD); //Not sure this is kosher but there will never be more than one impl of nsIVxmlPropertyDescriptor
  VxmRc rc = m_pVxm->setPropertyDescriptor(pd->m_pPD);
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE ;
}

NS_IMETHODIMP nsVxmlInterpreter::GetPropertyDescriptor(nsIVxmlPropertyDescriptor **_retval)
{
  NS_ADDREF(*_retval = new nsVxmlPropertyDescriptor(m_pVxm->getPropertyDescriptor()));
  return NS_OK;
}

NS_IMETHODIMP nsVxmlInterpreter::Allocate(PRInt16 *_retval)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::Allocate)
  *_retval = m_pVxm->allocate();
  printf("vxm allocate returned %d\n", *_retval);
  return NS_OK;
}

NS_IMETHODIMP nsVxmlInterpreter::Deallocate(PRInt16 *_retval)
{
   FUNCTION_CLASS_TRACE (nsVxmlInterpreter::Deallocate)
 *_retval = m_pVxm->deallocate();
  printf("vxm deallocate returned %d\n", *_retval);
  return NS_OK;
}

NS_IMETHODIMP 
nsVxmlInterpreter::InitDocumentScope(nsIDocument *aDoc)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::InitDocumentScope);

  // This handler starts the process to initialize a document scope (JavaScript) 
  // when a new document is loaded or a document in a tab receives focus.
  // We post an event to the sync queue which then waits on doc changing in case,
  // for example, a previous unload is still processing. It creates another event 
  // for the UI thread to set the current document. (This last was required because 
  // nsIDocument is not thread safe and various methods of executing synchronous operations
  // on the UI thread from a Monitored nsIEventQueue handler were not working.)
  //
  // UI Thread                  |Sync Thread                      |Speech Thread
  // ---------------------------|---------------------------------|--------------
  // InitDocumentScope()        |                                 |
  //   Post event   ----------->| HandleInitDocScopeEvent()       |
  //                            |    Wait on DocChanging          |
  //                            |    Enter Monitor                |              
  //                            |    Set DocChanging true         |
  //                            |    Post event to                |
  // SetCurrentDoc() <--------- |----- set CurrentDoc             |                    
  //   Enter Monitor            |                                 |                   
  //   Update current doc       |                                 |                
  //   Init MozillaScope        |                                 |                       
  //   Set DocChanging false    |                                 |                                                               
  //   Notify WaitOnDocChanging |                                 |
  //   Exit Monitor             |                                 |
  //                            |                                 |

  //Unecessary to initialize if we haven't gone to another doc 
  //e.g. when called by document focus right after load
  if(aDoc == mCurrentDoc)
    return NS_OK;

  //Put an event on our queue to do the work
  DocEvent * evt = new DocEvent(aDoc);
  PL_InitEvent(evt, this, nsVxmlInterpreter::HandleInitDocScopeEvent, nsVxmlInterpreter::DestroyInitDocScopeEvent);
  nsresult rv = mEventQ->PostEvent(evt);
  return NS_OK;
}

NS_IMETHODIMP 
nsVxmlInterpreter::AddHandler(const PRUnichar *aText, const PRUnichar *aDocURI, PRInt32 aTimeStamp, const PRUnichar *aBaseUri, const PRUnichar *aHandlerId, PRInt16 *_retval)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::AddHandler);

  // This function the process to add a handler to the VXM interpreter. It is called
  // by nsVxmlFormElement in its DOMContentLoaded handler or by XUL when adding a VoiceXML form
  // to speech enable the UI. We post an event to the sync thread which then waits on previous doc changing,
  // for example, InitDocumentScope. It then makes a synchronous call the the VXM interpreter to add the
  // handler. The VXM posts events to the UI thread to execute JS through MozillaScope.
  //
  // UI Thread                  |Sync Thread                      |Speech Thread
  // ---------------------------|---------------------------------|--------------
  // AddHandler()               |                                 |
  //    Post Event ------------>|HandleAddHandler()               |
  //                            |   Wait on DocChanging           |
  //                            |   AddHandler()                  |
  //                            |      MozillaScope posts events  |
  // Various JS_api() <---------|----- to execute JS              |
  //                            |                                 |
  //                            |                                 |

  //Put an event on our queue to do the work
  AddHandlerEvent * evt = new AddHandlerEvent(aText, aDocURI, aTimeStamp, aBaseUri, aHandlerId);
  PL_InitEvent(evt, this, nsVxmlInterpreter::HandleAddHandlerEvent, nsVxmlInterpreter::DestroyAddHandlerEvent);
  nsresult rv = mEventQ->PostEvent(evt);
 
  if (NS_FAILED(rv)) {
    PL_DestroyEvent(evt);
  }

  return NS_OK;
}

NS_IMETHODIMP 
nsVxmlInterpreter::RunHandler(nsIVxmlFormElement *aForm, const PRUnichar *aHandlerId, const PRUnichar *aDocUri, PRBool bDebugMode, PRInt16 *_retval)
{
  FUNCTION_CLASS_TRACE (nsIVxmlInterpreter::RunHandler);

  //
  // UI Thread                  |Sync Thread                      |Speech Thread
  // ---------------------------|---------------------------------|-----------------------------
  // RunHandler()               |                                 |
  //    PostEvent ------------->|HandleStartRunHandlerEvent()     |
  //                            |   Wait on DocChanging           |
  //                            |   start RunHandleRunnable ----->|RunHandleRunnable::Run
  //                            |                                 |   VxmInterpreter::runHandler()       
  //                            |                                 |      MozillaScope posts events 
  // Various JS_api() <---------|---------------------------------|----- to execute JS          
  //                            |                                 |   Enter DocChanging Monitor
  // HandleDispatchVxmlEvent()<-|-------------------------------- |-- Dispatch "vxmldone" XML event 
  //                            |                                 |      if doc not changing 
  //                            |                                 |   Exit DocChanging Monitor

  mRunHandlerRunnable = new RunHandlerRunnable(this, aHandlerId, aDocUri, bDebugMode);

  PLEvent * evt = new PLEvent;
  PL_InitEvent(evt, this, nsVxmlInterpreter::HandleStartRunHandlerEvent, nsVxmlInterpreter::DestroyStartRunHandlerEvent);
  nsresult rv = mEventQ->PostEvent(evt);

  return NS_OK;
}

NS_IMETHODIMP
nsVxmlInterpreter::OnDocumentUnload(nsIDocument* aDoc)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::OnDocumentUnload );

  // This event handler starts the process to "unload" a document. It places an event 
  // on the Sync Thread queue which must wait for the state to change to stopped and then
  // notify those that are waiting for dcument change to complete.
  //
  // UI Thread                  |Sync Thread                      |Speech Thread
  // ---------------------------|---------------------------------|--------------
  // OnDocumentUnload()         |                                 |
  //   Enter Monitor            |                                 |
  //   set DocChanging true     |                                 |
  //   cancel vxml form     ----|-------------------------------->|cancel()
  //   Exit Monitor             |                                 |
  //   send finish evt      --->|HandleFinishDocUnload()          |
  //                            |   Enter Monitor                 |VxmStateChange()
  //                            |   Wait for cancel complete      |state change to 
  //                            |                            <----|stopped
  //                            |   set DocChanging false         |
  //                            |   Notify WaitOnDocChanging      |                                                               
  //                            |   Exit Monitor                  |


  PR_EnterMonitor(mDocChangingMonitor);
  printf("Enter mDocChangingMonitor in OnDocumentUnload.\n");

  mbDocChanging = PR_TRUE;
  mbDocUnloading = PR_TRUE;
  printf("Unloading document %0.8x\n", aDoc);
  mCurrentDoc = nsnull; //todo: is this the right place to do this??
  mLoadedDocuments.Remove(aDoc);
  //todo: Unregister listeners?
  PRInt16 rc = MOZVXML_RC_OK;
  PRInt32 state;
  GetState(&state);
  if(!(nsIVxmlInterpreter::STOPPED & state || nsIVxmlInterpreter::STOPPING & state))
  {
    Cancel(&rc);
  }

  PR_ExitMonitor(mDocChangingMonitor);
  printf("Exit mDocChangingMonitor in OnDocumentUnload.\n");

  //Create and post the event to finish on a seperate thread
  PLEvent * evt = new PLEvent;
  PL_InitEvent(evt, this, nsVxmlInterpreter::HandleFinishDocUnloadEvent, nsVxmlInterpreter::DestroyFinishDocUnloadEvent);
  nsresult rv = mEventQ->PostEvent(evt);
 
  if (NS_FAILED(rv)) {
    PL_DestroyEvent(evt);
  }

  return rc;
}

NS_IMETHODIMP 
nsVxmlInterpreter::ClearHandlers(PRInt16 *_retval)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::ClearHandlers)
  *_retval = m_pVxm->clearHandlers();
  printf("vxm clearHandlers returned %d\n", *_retval);
  return NS_OK;
}

NS_IMETHODIMP nsVxmlInterpreter::SetUriResolver(nsIVxmlUriResolver *pUriResolver, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::GetUriResolver(nsIVxmlUriResolver **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::AddLogListener(nsIVxmlLogListener *pLogListener, PRInt16 logLevel, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::RemoveLogListener(nsIVxmlLogListener *pLogListener, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::Cancel(PRInt16 *_retval)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::Cancel)

  VxmRc rc = m_pVxm->cancel();
  *_retval = rc;
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE;
}

NS_IMETHODIMP nsVxmlInterpreter::StartListening(PRInt16 *_retval)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::StartListening)

  VxmRc rc = m_pVxm->startListening();
  *_retval = rc;
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE;
}

NS_IMETHODIMP nsVxmlInterpreter::StopListening(PRInt16 *_retval)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::StopListening)

  VxmRc rc = m_pVxm->stopListening();
  *_retval = rc;
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE;
}

NS_IMETHODIMP nsVxmlInterpreter::GetState(PRInt32 *mask)
{
 
  *mask = m_pVxm->getState();
  return MOZVXML_RC_OK;
}

NS_IMETHODIMP nsVxmlInterpreter::GetRecording(const PRUnichar *pstrRecHandle, PRUnichar **pstrFormat, nsIVxmlAudioBuffer **ppBuffers, PRUint16 *puiCount, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::ReleaseRecording(const PRUnichar *pstrRecHandle, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::AddLink(const PRUnichar *pstrLinkText, const PRUnichar *pstrDocUri, PRInt32 lDocTimeStamp, const PRUnichar *pstrBaseUri, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::AddSync(const PRUnichar *baseURI, const PRUnichar *pstrCurDocURI, const PRUnichar *pstrHtmlFormId, const PRUnichar *pstrFieldId, const PRUnichar *pstrHtmlName, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::SyncBlur(const PRUnichar *pstrHtmlFormId, const PRUnichar *pstrHtmlName, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::SyncFocus(const PRUnichar *pstrHtmlFormId, const PRUnichar *pstrHtmlName, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::Cancel_handler(const PRUnichar *pstrHandlerId, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::SetMaxInstances(PRInt16 iMaxInstances, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::GetMaxInstances(PRInt16 *_retval)
{
  *_retval = 69;
  return NS_OK;
}

NS_IMETHODIMP nsVxmlInterpreter::GetVersionString(PRUnichar **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::GetBuildString(PRUnichar **_retval)
{
  
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlInterpreter::GetUserAgentString(PRUnichar **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}


nsVxmlFeatureDescriptor::nsVxmlFeatureDescriptor()
{
  /* member initializers and constructor code */
}

nsVxmlFeatureDescriptor::~nsVxmlFeatureDescriptor()
{
  /* destructor code */
}

NS_IMETHODIMP nsVxmlFeatureDescriptor::CreateInstance(nsIVxmlFeatureDescriptor **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlFeatureDescriptor::SetValidation(PRBool bValidation)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlFeatureDescriptor::GetValidation(PRBool *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlFeatureDescriptor::SetCaching(PRBool bEnabled, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlFeatureDescriptor::GetCaching(PRBool *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

nsVxmlPropertyDescriptor::nsVxmlPropertyDescriptor()
{
  /* member initializers and constructor code */
}

nsVxmlPropertyDescriptor::~nsVxmlPropertyDescriptor()
{
  /* destructor code */
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::CreateInstance(nsIVxmlPropertyDescriptor **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::SetLocale(const PRUnichar *pstrLocale, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::GetLocale(PRUnichar **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::SetListeningMode(PRInt16 listeningMode, PRInt16 *_retval)
{
  VxmRc rc = m_pPD->setListeningMode((VxmListeningMode)listeningMode);
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE ;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::GetListeningMode(PRInt16 *_retval)
{
  VxmRc rc = (VxmRc) m_pPD->getListeningMode();
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE ;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::SetCacheSize(PRInt16 iCacheSize, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::GetCacheSize(PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::SupportedLocale(const PRUnichar *pstrLocale, PRBool *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::LocaleSwitchRequired(const PRUnichar *pstrLocale, PRBool *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::SetInstallPath(const PRUnichar *pstrPath, PRInt16 *_retval)
{
  VxmRc rc = m_pPD->setInstallPath(pstrPath);
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE ;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::GetInstallPath(PRUnichar **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::SetResourcePath(const PRUnichar *pstrPath, PRInt16 *_retval)
{
  VxmRc rc = m_pPD->setResourcePath(pstrPath);
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE ;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::GetResourcePath(PRUnichar **_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::SetProperty(PRInt16 name, PRInt16 value, PRInt16 *_retval)
{
  VxmRc rc = m_pPD->setProperty((VxmVoiceParameter)name, value);
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE ;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::GetProperty(PRInt16 name, PRInt16 *value, PRInt16 *_retval)
{
  VxmRc rc = m_pPD->getProperty((VxmVoiceParameter)name, (int *)value);
  return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE ;
}

NS_IMETHODIMP nsVxmlPropertyDescriptor::QuerySupportedPropertyValue(PRInt16 name, PRUnichar **value, PRInt16 *arraySize, PRBool bCTTS, PRInt16 *_retval)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

//*********************************************************************************************************
// nsIDOMEventListener methods

NS_IMETHODIMP 
nsVxmlInterpreter::HandleEvent(class nsIDOMEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::HandleEvent )

  nsAutoString s;
  PRUint16 phase = 0;
  aEvent->GetType(s);
  aEvent->GetEventPhase(&phase);

  wprintf(L"nsVxmlInterpreter::HandleEvent event type: %s phase: %u\n", s.get(), phase);

  // Don't handle this event unless the doc is the target
  if(!(phase == nsIDOMEvent::AT_TARGET))
  {
    return NS_OK;
  }

  //Get the target doc
  nsCOMPtr<nsIDOMEventTarget> tgt;
  aEvent->GetTarget(getter_AddRefs(tgt));
  nsCOMPtr<nsIDocument> aDoc = do_QueryInterface(tgt);

  NS_ASSERTION(mCurrentDoc, "Bad document");

  if ( s == NS_LITERAL_STRING("blur") )
  {
    return OnDocumentBlur(aDoc);
  }
  else if( s == NS_LITERAL_STRING("focus") )
  {
    return OnDocumentFocus(aDoc);
  }
  else if( s == NS_LITERAL_STRING("unload") )
  {
    return OnDocumentUnload(aDoc);
  }
  else
  {
    return NS_ERROR_UNEXPECTED;
  }
}

void 
nsVxmlInterpreter::WaitOnStateChange(PRInt32 aState)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::WaitOnStateChange);

  PR_EnterMonitor(mStateChangeMonitor);
  printf("Enter mStateChangeMonitor Waiting on state %x current state %x\n", aState, mState);
  while(!(aState & mState))
  {
    printf("Waiting on state %x current state %x\n", aState, mState);
    PR_Wait(mStateChangeMonitor, PR_INTERVAL_NO_TIMEOUT);
  }
  printf("Wait over. State changed to %x\n", aState);

  PR_ExitMonitor(mStateChangeMonitor);
  printf("Exit  mStateChangeMonitor\n");
}

void 
nsVxmlInterpreter::WaitOnDocChanging()
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::WaitOnDocChanging);

  PR_EnterMonitor(mDocChangingMonitor);
  printf("Enter mDocChangingMonitor in WaitOnDocChanging.\n");

  while(mbDocChanging)
  {
    printf("Waiting on document change.\n");
    PR_Wait(mDocChangingMonitor, PR_INTERVAL_NO_TIMEOUT);
  }

  PR_ExitMonitor(mDocChangingMonitor);
  printf("Exit  mDocChangingMonitor in WaitOnDocChanging.\n");
}

//*********************************************************************************************************
// nsVxmlInterpreter::VxmStateListener methods

void 
nsVxmlInterpreter::vxmStateChange(long lPrevState, long lNewState)
{
  FUNCTION_CLASS_TRACE (nsVxmlInterpreter::vxmStateChange);
  printf("State change from %x to %x\n", lPrevState, lNewState);
  PR_EnterMonitor(mStateChangeMonitor);
  printf("Enter mStateChangeMonitor for State change from %x to %x\n", lPrevState, lNewState);
  mState = lNewState;
  PR_Notify(mStateChangeMonitor);

  //Notify nsIObservers here
  PR_EnterMonitor(mDocChangingMonitor);
  printf("Enter mDocChangingMonitor in vxmStateChange\n");

  // I would rather do this by observing "xpcom-shutdown" but that comes too late
  // to prevent state change events in flight following the cancel. So we prevent notification
  // during any doc unloading to prevent calling the observers after a document is destroyed.
  if(!mbDocUnloading)
    NotifyStateChangeObservers(lNewState);
  PR_ExitMonitor(mDocChangingMonitor);
  printf("Exit mDocChangingMonitor in vxmStateChange\n");

  PR_ExitMonitor(mStateChangeMonitor);
  printf("Exit mStateChangeMonitor\n");

 

}

void 
nsVxmlInterpreter::vxmError(VxmRc rc){};


//*********************************************************************************************************
// VxmXmlEventListener methods

void 
nsVxmlInterpreter::vxmXmlEvent(const wchar_t *pstrElementType, const wchar_t *pstrElementId, const wchar_t *pstrEvent)
{
  wprintf(L"!!!!!!!!!!!!!!!!!!!!!!!VXML event: Element type = \"%s\"\tElement id =\"%s\"\tEvent = \"%s\"\n", pstrElementType, pstrElementId, pstrEvent);
  nsAutoString id(pstrElementId);
  nsAutoString event(pstrEvent);
  DispatchVxmlEvent(id, event);
}


//*********************************************************************************************************
// nsVxmlInterpreter private methods

void 
nsVxmlInterpreter::GetWindowFromDocument(nsIDOMDocument* aDocument, nsIDOMWindowInternal** aWindow)
{
  nsCOMPtr<nsIDOMDocumentView> docview = do_QueryInterface(aDocument);

  nsCOMPtr<nsIDOMAbstractView> view;
  docview->GetDefaultView(getter_AddRefs(view));
  if (!view) return;
 
  nsCOMPtr<nsIDOMWindowInternal> window = do_QueryInterface(view);
  NS_IF_ADDREF(*aWindow = window);
}

NS_IMETHODIMP
nsVxmlInterpreter::OnDocumentBlur(nsIDocument* aDoc)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::OnDocumentBlur )
  PRInt16 rc;
  Cancel(&rc);
  return rc;
}

NS_IMETHODIMP
nsVxmlInterpreter::OnDocumentFocus(nsIDocument* aDoc)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::OnDocumentFocus )
  if (aDoc) //why is this being called with null document?!!
    return InitDocumentScope(aDoc);
  else
    return NS_OK;
}

void
nsVxmlInterpreter::RegisterEventListeners(nsIDocument* aDoc)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::RegisterEventListeners );

  //Registering for the document events
  nsCOMPtr<nsIDOMDocument> domdoc;
  domdoc = do_QueryInterface(aDoc);
  nsCOMPtr<nsIDOMWindowInternal> domWindow;
  GetWindowFromDocument(domdoc, getter_AddRefs(domWindow));
  nsCOMPtr<nsIDOMEventTarget> eventTarget = do_QueryInterface(domWindow);
  NS_ASSERTION(eventTarget, "Bad document event target");

  eventTarget->AddEventListener(NS_LITERAL_STRING("blur"), this, PR_TRUE);
  eventTarget->AddEventListener(NS_LITERAL_STRING("focus"), this, PR_TRUE);
  eventTarget->AddEventListener(NS_LITERAL_STRING("unload"), this, PR_TRUE);
}

void
nsVxmlInterpreter::SetCurrentDoc(nsIDocument * aDoc)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::SetCurrentDoc );

  PR_EnterMonitor(mDocChangingMonitor);
  printf("Enter mDocChangingMonitor in SetCurrentDoc\n");

  mCurrentDoc = aDoc;

  //If this is a new document add to the list, initialize scope
  //and register listeners on the doc
  PRInt32 data;
  PRBool bIsNewDoc = !mLoadedDocuments.Get(mCurrentDoc, &data);
  if(bIsNewDoc)
  {
    //Only storing the doc "index" for now...
    mLoadedDocuments.Put(mCurrentDoc, mLoadedDocuments.Count());

    // Initialize JavaScript scope using the document
    MozillaScope::Init(mCurrentDoc);

    RegisterEventListeners(mCurrentDoc);
  }

  // Get the global object and native context from the Document 
  nsIScriptGlobalObject *sgo = mCurrentDoc->GetScriptGlobalObject();
  JSObject * glob = sgo->GetGlobalJSObject();
  nsIScriptContext *scrCx = sgo->GetContext();
  JSContext* cx = (JSContext *) scrCx->GetNativeContext();

  printf("New document 0x%0.8x global object 0x%0.8x and context 0x%0.8x\n", mCurrentDoc, sgo, cx);

  //Set the global object, context, and scope callback in the Interpreter
  VxmRc rc;
  rc = m_pVxm->setECMAScriptContext(cx);

  //Use the nsIDocument as the global scope
  rc = m_pVxm->setECMAScriptScope(mCurrentDoc);
  m_pVxm->setScopeFactory(MozillaScope::createScope);

  //All finished, we can depart
  mbDocChanging = PR_FALSE;
  PR_Notify(mDocChangingMonitor);
  printf("Notified mDocChangingMonitor\n");
  PR_ExitMonitor(mDocChangingMonitor);
  printf("Exit mDocChangingMonitor in SetCurrentDoc\n");
}

void 
nsVxmlInterpreter::DispatchVxmlEvent(const nsString& aFormId, const nsString& aEventName)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::DispatchVxmlEvent );

  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  NS_ASSERTION(eqsvc, "Failed to get nsIEventQueueService.");

  nsCOMPtr<nsIEventQueue> eventQ;
  eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                              getter_AddRefs(eventQ));
  NS_ASSERTION(eventQ, "Failed to get UI Event Queue.");

  DispatchVxmlEventEvent * evt = new DispatchVxmlEventEvent(aFormId, aEventName);
  if (!evt) 
  {
     return;
  }
 
  PL_InitEvent(evt, this, nsVxmlInterpreter::HandleDispatchVxmlEvent, nsVxmlInterpreter::DestroyDispatchVxmlEvent);
 
  void * result;
  nsresult rv = eventQ->PostSynchronousEvent(evt, &result);
 
   if (NS_FAILED(rv)) {
     PL_DestroyEvent(evt);
   }
   return;
}

void *
nsVxmlInterpreter::HandleDispatchVxmlEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::HandleDispatchVxmlEvent );

  DispatchVxmlEventEvent * evt = NS_STATIC_CAST(DispatchVxmlEventEvent *, aEvent);
  nsVxmlInterpreter * vxml = NS_STATIC_CAST(nsVxmlInterpreter *, aEvent->owner);

  NS_ASSERTION(vxml->mCurrentDoc, "Invalid current doc in VxmlIntepreter.");

  //Create the event
  nsCOMPtr<nsIDOMDocumentEvent> docEvent = do_QueryInterface(vxml->mCurrentDoc);
  NS_ASSERTION(docEvent, "Invalid nsIDOMDocumentEvent.");

  PRBool cancelable = PR_TRUE;
  PRBool bubbles = PR_TRUE;
  nsCOMPtr<nsIDOMEvent> event;
  docEvent->CreateEvent(NS_LITERAL_STRING("Events"), getter_AddRefs(event));
  event->InitEvent(evt->mEventName, bubbles, cancelable);

  //Get the target element
  nsCOMPtr<nsIDOMDocument> domdoc = do_QueryInterface(vxml->mCurrentDoc);
  nsCOMPtr<nsIDOMElement> form;
  domdoc->GetElementById(evt->mFormId, getter_AddRefs(form));


  //nsCOMPtr<nsIDOMEventTarget> target = do_QueryInterface(vxml->mCurrentDoc); //using doc as target
  nsCOMPtr<nsIDOMEventTarget> target = do_QueryInterface(form); // using vxml Form as target
  NS_ASSERTION(target, "Failed to obtain event target.");
  PRBool defaultActionEnabled;
  target->DispatchEvent(event, &defaultActionEnabled);
  return aEvent;
}

void
nsVxmlInterpreter::DestroyDispatchVxmlEvent(PLEvent *evt) {/* delete evt; */}

void 
nsVxmlInterpreter::NotifyStateChangeObservers(PRInt32 aNewState)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::NotifyStateChangeObservers );

  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  NS_ASSERTION(eqsvc, "Failed to get nsIEventQueueService.");

  nsCOMPtr<nsIEventQueue> eventQ;
  eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                              getter_AddRefs(eventQ));

  NS_ASSERTION(eventQ, "Failed to get Event Queue.");

  NotifyStateChangeObserversEvent * evt = new NotifyStateChangeObserversEvent(aNewState);
  NS_ASSERTION(evt, "Failed to create event.");
 
  PL_InitEvent(evt, this, nsVxmlInterpreter::HandleNotifyStateChangeObserversEvent, nsVxmlInterpreter::DestroyNotifyStateChangeObserversEvent);
 
  nsresult rv = eventQ->PostEvent(evt);
 
  if (NS_FAILED(rv)) {
    PL_DestroyEvent(evt);
  }
  return;
}

void *
nsVxmlInterpreter::HandleNotifyStateChangeObserversEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::HandleNotifyStateChangeObserversEvent );

  NotifyStateChangeObserversEvent * evt = NS_STATIC_CAST(NotifyStateChangeObserversEvent *, aEvent);
  nsVxmlInterpreter * self = NS_STATIC_CAST(nsVxmlInterpreter *, aEvent->owner);


  //Don't notify observers if we've started shutting down
  if(self->mbShuttingDown)
    return nsnull;

  nsCOMPtr<nsIObserverService> os = do_GetService("@mozilla.org/observer-service;1");
  NS_ASSERTION(os, "Failed to get nsIObserverService.");

  nsCOMPtr<nsISupportsPRInt32> newState = do_CreateInstance(NS_SUPPORTS_PRINT32_CONTRACTID);
  NS_ASSERTION(newState, "Couldn't create nsISupportsPRInt32");

  newState->SetData(evt->mNewState);
  os->NotifyObservers(newState, "notifyOnVxmlStateChange", nsnull);

  return aEvent;
}

void
nsVxmlInterpreter::DestroyNotifyStateChangeObserversEvent(PLEvent *evt) {/* delete evt; */}

void * 
nsVxmlInterpreter::HandleFinishDocUnloadEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::HandleFinishDocUnloadEvent );

  nsVxmlInterpreter * self = NS_STATIC_CAST(nsVxmlInterpreter *, aEvent->owner);

  //Any running form was cancelled. This wait is redundant with the join below
  self->WaitOnStateChange(VxmInterpreter::STOPPED);

  //We want to serialize doc unload and initdoc so we will wait for the
  //RunHandlerThread to complete in a critical section.
  PR_EnterMonitor(self->mDocChangingMonitor);
  printf("Enter mDocChangingMonitor in HandleFinishDocUnloadEvent.\n");
 
  //Remove the handlers
  PRInt16 rv;
  self->ClearHandlers(&rv);

  //Block until RunHandler returns then clean it up
  //PR_JoinThread(self->mRunHandlerThread);
  //  delete self->mRunHandlerThread;
  self->mRunHandlerThread = nsnull;

  //All done, set the flag and depart
  self->mbDocChanging = PR_FALSE;
  self->mbDocUnloading = PR_FALSE;
  PR_Notify(self->mDocChangingMonitor);
  PR_ExitMonitor(self->mDocChangingMonitor);
  printf("Exit  mDocChangingMonitor in HandleFinishDocUnloadEvent.\n");
  return aEvent;
}

void 
nsVxmlInterpreter::DestroyFinishDocUnloadEvent(PLEvent *aEvent) {delete aEvent;}

void * 
nsVxmlInterpreter::HandleInitDocScopeEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::HandleInitDocScopeEvent );

  DocEvent * evt = NS_STATIC_CAST(DocEvent *, aEvent);
  nsVxmlInterpreter * self = NS_STATIC_CAST(nsVxmlInterpreter *, aEvent->owner);

  //Initializing a document must be serialized after unloading a document
  //if a document is being unloaded.
  self->WaitOnDocChanging();

  //Need to serialize this to prevent this doc being unloaded before its finished loading
  PR_EnterMonitor(self->mDocChangingMonitor);
  printf("Enter 1 mDocChangingMonitor in HandleInitDocScopeEvent\n");
  self->mbDocChanging = PR_TRUE;

  PR_ExitMonitor(self->mDocChangingMonitor);
  printf("Exit 1 mDocChangingMonitor in HandleInitDocScopeEvent\n");

  //Send back to main thread
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  NS_ASSERTION(eqsvc, "Failed to get nsIEventQueueService.");

  nsCOMPtr<nsIEventQueue> eventQ;
  eqsvc->GetSpecialEventQueue(nsIEventQueueService::UI_THREAD_EVENT_QUEUE,
                              getter_AddRefs(eventQ));
  NS_ASSERTION(eventQ, "Failed to get UI Event Queue.");
  DocEvent * newevt = new DocEvent(evt->mDoc);
  PL_InitEvent(newevt, self, nsVxmlInterpreter::HandleSetCurrentDocEvent, nsVxmlInterpreter::DestroySetCurrentDocEvent);
  eventQ->PostEvent(newevt);
  return aEvent;
}

void 
nsVxmlInterpreter::DestroyInitDocScopeEvent(PLEvent *aEvent) {delete aEvent;}

void * 
nsVxmlInterpreter::HandleAddHandlerEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::HandleAddHandlerEvent );

  AddHandlerEvent * evt = NS_STATIC_CAST(AddHandlerEvent *, aEvent);
  nsVxmlInterpreter * self = NS_STATIC_CAST(nsVxmlInterpreter *, aEvent->owner);

  self->WaitOnDocChanging();
  VxmRc rc = self->m_pVxm->addHandler(evt->mText.get(), evt->mDocURI.get(), evt->mDocTimeStamp, evt->mBaseURI.get(), evt->mHandlerId.get());
  printf ("Add Handler returned %d\n", rc);
  return aEvent;
}

void 
nsVxmlInterpreter::DestroyAddHandlerEvent(PLEvent *aEvent){ delete aEvent; }

void * 
nsVxmlInterpreter::HandleStartRunHandlerEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::HandleStartRunHandlerEvent )

  nsVxmlInterpreter * self = NS_STATIC_CAST(nsVxmlInterpreter *, aEvent->owner);

  self->WaitOnDocChanging();

  nsresult rv = NS_NewThread(getter_AddRefs(self->mCurrentRunHandlerThrd),
                             self->mRunHandlerRunnable,
                             0,
                             PR_JOINABLE_THREAD);

  return aEvent; 
}

void 
nsVxmlInterpreter::DestroyStartRunHandlerEvent(PLEvent *aEvent){/*todo:*/}

void * 
nsVxmlInterpreter::HandleSetCurrentDocEvent(PLEvent *aEvent)
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::HandleSetCurrentDocEvent );

  DocEvent * evt = NS_STATIC_CAST(DocEvent *, aEvent);
  nsVxmlInterpreter * self = NS_STATIC_CAST(nsVxmlInterpreter *, aEvent->owner);
  self->SetCurrentDoc(NS_STATIC_CAST(nsIDocument *, evt->mDoc));
  return aEvent;
}

void 
nsVxmlInterpreter::DestroySetCurrentDocEvent(PLEvent *aEvent) { delete aEvent; }

void PR_CALLBACK 
nsVxmlInterpreter::ptt(void * arg)
{
  nsVxmlInterpreter * pVxml = (nsVxmlInterpreter *)arg;

  int c;
  PRInt16 rc;

  while (PR_TRUE)
  {
    c = getchar();
    pVxml->mbStartListening = ! pVxml->mbStartListening;
    if(pVxml->mbStartListening)
    {
      pVxml->StartListening(&rc);
    }
    else
    {
       pVxml->StopListening(&rc);
    }
    printf("Start listening changed to %s\n", pVxml->mbStartListening ?  "true" : "false");
  }
}

// Thread for calling the RunHandler
void PR_CALLBACK 
nsVxmlInterpreter::RunHandlerThread(void * arg) 
{
  FUNCTION_CLASS_TRACE ( nsVxmlFormElement::RunHandlerThread )

  RunHandlerArgs * args = (RunHandlerArgs *)arg;
  VxmInterpreter * vxm = args->mInterpreter->m_pVxm;
  VxmRc rc = vxm->runHandler(args->mHandlerId.get(), args->mDocURI.get(), args->mDebug);

  wprintf(L"RunHandler return: %d\n", rc);

  if(rc == VXM_RC_OK)
  {
    //Don't dispatch the event if the document is unloading
    //This can happen if "location.href" is modified in a VoiceXML script
    PR_EnterMonitor(args->mInterpreter->mDocChangingMonitor);
    printf("Enter mDocChangingMonitor in RunHandlerThread\n");

    if(!args->mInterpreter->mbDocChanging)
    {
      printf("Dispatch vxmldone from nsVxmlInterpreter::RunHandlerThread\n");
      args->mInterpreter->DispatchVxmlEvent(args->mHandlerId, NS_LITERAL_STRING("vxmldone"));
    }

    PR_ExitMonitor(args->mInterpreter->mDocChangingMonitor);
    printf("Exit mDocChangingMonitor in RunHandlerThread\n");

  }
  //Need a scheme to process the error paths return rc == VXM_RC_OK ? NS_OK : NS_ERROR_FAILURE;
}

//*********************************************************************************************************
// nsVxmlLogListener methods

nsVxmlLogListener::nsVxmlLogListener()
{
  /* member initializers and constructor code */
}

nsVxmlLogListener::~nsVxmlLogListener()
{
  /* destructor code */
}

NS_IMETHODIMP nsVxmlLogListener::VxmLogEntry(const PRUnichar *pstrLogMessage, PRInt16 logLevel, PRInt32 secs, PRInt16 iMillisecs)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

//*********************************************************************************************************
// nsIObserver


NS_IMETHODIMP
nsVxmlInterpreter::Observe( nsISupports *aSubject, const char *aTopic, const PRUnichar *someData ) 
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::Observe );

  nsCString topic( aTopic );
  printf("Observed %s event.\n", aTopic);
  if (topic.EqualsLiteral("xpcom-shutdown"))
  {
    //Set the flag. Only accessed on the UI thread so no need for monitor
    mbShuttingDown = PR_TRUE;
    MozillaScope::Shutdown();
  }

  return NS_OK;
}

//*********************************************************************************************************
// nsIRunnable - implements Event Queue

NS_IMETHODIMP 
nsVxmlInterpreter::Run()
{
  FUNCTION_CLASS_TRACE ( nsVxmlInterpreter::Run );

  PR_EnterMonitor(mEventQMonitor);
  nsCOMPtr<nsIEventQueueService> eqsvc = do_GetService("@mozilla.org/event-queue-service;1");
  NS_ASSERTION(eqsvc, "Failed to get nsIEventQueueService.");

  eqsvc->CreateMonitoredThreadEventQueue();
  eqsvc->GetThreadEventQueue(NS_CURRENT_THREAD, getter_AddRefs(mEventQ));
  NS_ASSERTION(mEventQ, "Failed to create event queue.");

  PR_Notify(mEventQMonitor);
  PR_ExitMonitor(mEventQMonitor);

  //Event Loop
  PLEvent *evt;
  while(PR_TRUE)
  {
    mEventQ->WaitForEvent(&evt);
    printf("Handle event %x\n", evt);
    mEventQ->HandleEvent(evt);
    printf("Return from handle event %x\n", evt);
  }

  return NS_OK;
}

NS_IMETHODIMP 
RunHandlerRunnable::Run()
{
  FUNCTION_CLASS_TRACE ( RunHandlerRunnable::Run );

  VxmRc rc = mInterpreter->m_pVxm->runHandler(mHandlerId.get(), mDocURI.get(), mDebug);

  wprintf(L"RunHandler return: %d\n", rc);

  //Don't dispatch the vxmldone event if the document is unloading
  //This can happen if "location.href" is modified in a VoiceXML script
  PR_EnterMonitor(mInterpreter->mDocChangingMonitor);
  printf("Enter mDocChangingMonitor in RunHandlerThread\n");

  if(mInterpreter->mCurrentDoc && !mInterpreter->mbDocChanging && rc == VXM_RC_OK)
    {
      printf("Dispatch vxmldone from nsVxmlInterpreter::RunHandlerThread\n");
      mInterpreter->DispatchVxmlEvent(mHandlerId, NS_LITERAL_STRING("vxmldone"));
    }

  PR_ExitMonitor(mInterpreter->mDocChangingMonitor);
  printf("Exit mDocChangingMonitor in RunHandlerThread\n");

  return NS_OK;
}
